(:
Copyright (C) 2011 Adam Retter <adam.retter@googlemail.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
:)
module namespace semantic = "http://www.adamretter.org.uk/semantic";

import module namespace http = "http://expath.org/ns/http-client";

declare namespace xh = "http://www.w3.org/1999/xhtml";

declare namespace rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#";
declare namespace j.3="http://fise.iks-project.eu/ontology/";
declare namespace j.1="http://purl.org/dc/terms/";

declare variable $semantic:stanbol-engines-api := "http://localhost:9090/engines";

declare variable $semantic:rdf-xml-mimetype := "application/rdf+xml";
declare variable $semantic:xhtml-mimetype := "application/xhtml+xml";

declare variable $semantic:web-rest-api := "http://localhost:8080/exist/rest";
declare variable $semantic:css := xs:anyURI("/db/enrichment/semantic.css");

declare function semantic:annotate($url as xs:string, $content-in as document-node()) {
        
    (: TODO - stack overflow error check :)
    let $stored-content-uri := xmldb:store("/db/enrichment/cache/content", (), $content-in, $semantic:xhtml-mimetype),
    $content := fn:doc($stored-content-uri) return
    (: end TODO :)
   
        (: enhance the body text :)
        let $body-text-content := $content/*:html/*:body//text()[not(parent::*:script)] return        
            let $enhanced-body-text-content-in := semantic:stanbol-enhance($url, $body-text-content),
            $db-uri-enhanced-content := xmldb:store("/db/enrichment/cache/enhanced", (), $enhanced-body-text-content-in, $semantic:rdf-xml-mimetype),
            $enhanced-body-text-content := doc($db-uri-enhanced-content) return

        let $text-nodes-to-enhance := semantic:match-text-nodes-to-enhance($body-text-content, $enhanced-body-text-content) return
            
            semantic:annotate-content($content, $content, $text-nodes-to-enhance, $db-uri-enhanced-content, $enhanced-body-text-content)
            
        (:
       $enhanced-content := semantic:annotate-content($content, $db-uri-enhanced-content, $enhanced-body-text-content) return
       :)
};

declare function semantic:match-text-nodes-to-enhance($body-text-content as text()*, $enhanced-body-text-content as document-node()) as element(match)* {
    let $descriptions-with-start := 
        for $desc in $enhanced-body-text-content//rdf:Description[j.3:start] order by xs:int($desc/j.3:start/text()) return
            $desc
    return
        semantic:find-text-nodes($descriptions-with-start, 0, $body-text-content)
};

declare function semantic:find-text-until($current-offset as xs:int, $end as xs:int, $list as text()*) as text()* {
	if(empty($list))then
		()
	else
		let $text := $list[1] return
			if($current-offset < $end)then
			(
				$text,
				semantic:find-text-until($current-offset + string-length($text), $end, subsequence($list, 2))
			)
			else()
};

declare function semantic:find-text-nodes($descriptions as element(rdf:Description)*, $current-offset as xs:int, $list as text()*) as element(match)* {
    if (empty($list) or empty($descriptions)) then
        ()
    else
        let $text := $list[1]
        let $end := $current-offset + string-length($text)
		
		let $description-start := xs:int($descriptions[1]/j.3:start/text())
		let $description-end := xs:int($descriptions[1]/j.3:end/text())
		let $text-start-offset := $description-start - $current-offset
		let $text-end-offset := $description-end - $current-offset
        return
			if($description-start < $end)then
				let $relevant-texts :=
					if($description-end > $end)then
						(
						  $text,
						  semantic:find-text-until($end, $description-end, fn:subsequence($list, 2))
						)
					else
						$text
				return
				(
				    (: useful for debugging :)
				    (: 
    				<match>
    				    <text-node>{$relevant-texts}</text-node>	
    					<portion currentOffset="{$currentOffset}" end="{$end}" startOffset="{$text-start-offset}" endOffset="{$text-end-offset}">{fn:substring(fn:string-join($relevant-texts, ""), $text-start-offset+1, $text-end-offset - $text-start-offset)}</portion>
    					{$descriptions[1]}
    				</match>,
    				:)
    				<match>
    					<texts start="{$text-start-offset}" end="{$text-end-offset}">
    					{
    						for $relevant-text in $relevant-texts return
    							<text node-id="{util:node-id($relevant-text)}"/>
    					}
    					</texts>
    					{
    					$descriptions[1]
    					}
    					<description node-id="{util:node-id($descriptions[1])}"/>
    				</match>
    				,
    				semantic:find-text-nodes(subsequence($descriptions, 2), $current-offset, $list)
				)
			else
				 semantic:find-text-nodes($descriptions, $end, subsequence($list, 2))
};

declare function semantic:inject-into-head($head as element(), $db-uris-enhanced-content as xs:anyURI*, $enhanced-body-text-content as document-node()) {
    
    let $link-qn := fn:QName(fn:namespace-uri-from-QName(fn:node-name($head)), fn:string-join((fn:prefix-from-QName(fn:node-name($head)), "link"),":")) return
    (
        comment {
            "AND THEN..."
        },
    
        (: inject in any alternative representations :)
        for $uri in $db-uris-enhanced-content return
            element { $link-qn } {
                attribute rel { "alternative" },
                attribute type { $semantic:rdf-xml-mimetype },
                attribute href { semantic:web-url-from-db-uri($uri) }
            }
        ,
        
        (: inject in semantic css display :)
        element { $link-qn} {
            attribute rel { "stylesheet" },
            attribute type { "text/css" },
            attribute href { semantic:web-url-from-db-uri($semantic:css) }
        },
        
        (: inject categories :)
        for $desc-category in $enhanced-body-text-content/rdf:RDF/rdf:Description[j.1:type/@rdf:resource = "http://fise.iks-project.eu/ontology/Category"] return
            <meta name="DC.subject" scheme="{$desc-category/j.3:entity-type/@rdf:resource}" content="{$desc-category/j.3:entity-label}"/>
        
    )
    (: //*:Description[*:type/@*:resource = "http://fise.iks-project.eu/ontology/Category"] :)
};

declare function semantic:inject-into-body($body as element(), $enhanced-body-text-content as document-node()) {
    let $div-qn := fn:QName(fn:namespace-uri-from-QName(fn:node-name($body)), fn:string-join((fn:prefix-from-QName(fn:node-name($body)), "div"),":")),
    $ol-qn := fn:QName(fn:namespace-uri-from-QName(fn:node-name($body)), fn:string-join((fn:prefix-from-QName(fn:node-name($body)), "ol"),":")),
    $li-qn := fn:QName(fn:namespace-uri-from-QName(fn:node-name($body)), fn:string-join((fn:prefix-from-QName(fn:node-name($body)), "li"),":")),
    $a-qn := fn:QName(fn:namespace-uri-from-QName(fn:node-name($body)), fn:string-join((fn:prefix-from-QName(fn:node-name($body)), "a"),":")),
    $span-qn := fn:QName(fn:namespace-uri-from-QName(fn:node-name($body)), fn:string-join((fn:prefix-from-QName(fn:node-name($body)), "span"),":")) return
    
        element { $div-qn } {
            attribute id { "enr-explicit-categories" },
            <p id="enr-explicit-categories-title">Found Categories</p>,
            
            element {$ol-qn} {
                (: inject categories :)
                for $desc-category in $enhanced-body-text-content/rdf:RDF/rdf:Description[j.1:type/@rdf:resource = "http://fise.iks-project.eu/ontology/Category"] order by xs:double($desc-category/j.3:confidence) descending return
                    element {$li-qn} {
                        attribute about { "" },
                        attribute class {"enr-subject"},
                        element {$span-qn} {
                            attribute property { "dc:subject" },
                            attribute type { $desc-category/j.3:entity-type/@rdf:resource },
                            attribute class {"enr-subject-label"},
                            element {$a-qn} {
                                attribute href { $desc-category/j.3:entity-reference/@rdf:resource },
                                attribute class { "enr-ignore" },
                                $desc-category/j.3:entity-label/text()
                            }
                        },
                        element {$span-qn} {
                            attribute class {"enr-subject-confidence"},
                            fn:concat(fn:round((xs:double($desc-category/j.3:confidence) div 1) * 100), "%")
                        }
                    }
            }
        }
        
                            (: <meta name="DC.subject" scheme="{$desc-category/j.3:entity-type/@rdf:resource}" content="{$desc-category/j.3:entity-label}"/> :)
};

declare function semantic:annotate-content($content as node(), $original-content-doc as document-node(), $text-nodes-to-enhance as element(match)*, $db-uris-enhanced-content as xs:anyURI*, $enhanced-body-text-content as document-node()) {
    typeswitch($content) 
        case $doc as document-node() return
            document {
                for $n in $doc/node() return
                    semantic:annotate-content($n, $original-content-doc, $text-nodes-to-enhance, $db-uris-enhanced-content, $enhanced-body-text-content)
            }
        case $elem as element() return
            element { fn:node-name($elem) } {
                
                (: add a profile to the head if it doesnt exist :)
                if(fn:lower-case(fn:local-name($elem)) eq "head")then
                    if(fn:empty($elem[@profile]))then
                        attribute profile { "http://dublincore.org/documents/dcq-html/" }
                    else()
                else(),
                
                (: process existing attributes :)
                $elem/attribute(),
                
                (: append to the body some explicit display semantics :)
                if(fn:lower-case(fn:local-name($elem)) eq "body")then
                    semantic:inject-into-body($elem, $enhanced-body-text-content)
                else(),
                
                (: process existing nodes :)
                for $n in $elem/node() return
                    semantic:annotate-content($n, $original-content-doc, $text-nodes-to-enhance, $db-uris-enhanced-content, $enhanced-body-text-content)
                ,
                
                (: append to the head some semantics :)
                if(fn:lower-case(fn:local-name($elem)) eq "head")then
                    semantic:inject-into-head($elem, $db-uris-enhanced-content, $enhanced-body-text-content)
                else()
            }
        
        
        case $text as text() return
            
            let $enhancements := $text-nodes-to-enhance[texts/text/@node-id = util:node-id($text)] return
            if(fn:empty($enhancements))then
                $text
            else
                semantic:enhance-text($original-content-doc, $text, $enhancements, $enhanced-body-text-content)
        
        default $x return
            $x
};

declare function semantic:enhance-text($original-content-doc as document-node(), $text as text(), $enhancements as element(match)+, $enhanced-body-text-content as document-node()) {
        for $enhancement in $enhancements
        let $first-text-node := $enhancement/texts/text[position() eq 1]
        let $last-text-node := $enhancement/texts/text[position() eq last()]
        let $found-first := xs:boolean($first-text-node/string(@node-id) eq util:node-id($text))
        let $found-last := xs:boolean($last-text-node/string(@node-id) eq util:node-id($text))
        return
        
        if($found-first eq false() and $found-last eq false())then
            (: skip middle entries these are processed by first entry :)
            ()
        else if($found-first eq false() and $found-last eq true())then
            (: skip last entry these are processed by the first entry :)
            ()
        else
            (: process first, and from here then also process middle and last entries :)
            let $texts := $first-text-node/parent::texts
            let $start-offset := xs:int($texts/@start)
            let $start := if($start-offset eq 0) then 1 else $start-offset
            let $end := xs:int($texts/@end)
            let $total := fn:count($texts/text)
            let $desc := util:node-by-id($enhanced-body-text-content, $texts/parent::match/description/@node-id)
            return
                
                let $txt := string-join(
                    for $t in $texts/text return
                        util:node-by-id($original-content-doc, $t/@node-id)
                    ," "
                )
                let $txt-length := fn:string-length($txt)
                return
                
                    if($start eq 1 and $txt-length eq $end) then
                            (: no split, all enhance  :)
                            semantic:rdfa-span($desc, $txt, "ne1")
                        else if($start eq 1 and $txt-length lt $end - $start)then
                            (: no split, no enhance :)
                            text { $txt }
                        else if($start eq 1 and $txt-length gt $end) then
                        (
                            (: split based on end :)
                            let $span-content := fn:substring($txt, 1, $end + ($total - 1)) return
                                semantic:rdfa-span($desc, $span-content, "ne2")
                            ,
                            text{fn:substring($txt, $end + ($total))}
                        )
                        else if($start ge 1 and $txt-length eq $end)then
                        (
                            (: split based on start :)
                            text{fn:substring($txt, 1, $start)},
                            let $span-content := fn:substring($txt, $start + 1) return
                                semantic:rdfa-span($desc, $span-content, "ne3")
                            
                        )
                        else
                        
                            (: split in middle :)
                            let $offset :=
                                if($total eq 1)then
                                    0
                                else
                                    $total + 1
                            let $offset2 :=
                                if($total eq 1)then
                                    0
                                else
                                    $total
                            return
                            (
                                text{fn:substring($txt, 1, $start - $offset)},
                                let $span-content := fn:substring($txt, $start - $offset + 1, $end - ($start - $offset) - $offset2) return
                                    semantic:rdfa-span($desc, $span-content, "ne4")
                                ,
                                text{fn:substring($txt, $end - $offset2 + 1)}
                            )
};

declare function semantic:rdfa-span($desc as element(rdf:Description), $content as xs:string, $classes as xs:string*) as element (xh:span) {
    element xh:span {
        
        attribute about { $desc/@rdf:about },
        if($desc/j.1:type)then
            attribute typeof { $desc/j.1:type/@rdf:resource }
        else(),
        attribute class {
            string-join(
                for $class in ("ne", $classes) return
                    $class
            , " ")
        },
        attribute style { "display : inline;" },
        
        text { $content }
    }
};

declare function semantic:web-url-from-db-uri($uri as xs:anyURI) as xs:anyURI {
    xs:anyURI(
        fn:concat($semantic:web-rest-api, $uri)
    )
};

declare function semantic:stanbol-enhance($url as xs:string, $text as text()*) as document-node() {
    let $req :=
        <http:request href="{$semantic:stanbol-engines-api}?uri={fn:encode-for-uri($url)}" method="post" http="1.0">
            <http:header name="Accept" value="{$semantic:rdf-xml-mimetype}"/>
            <http:body media-type="text/plain">
            { $text }
            </http:body>
        </http:request>
    return
        
        let $resp := http:send-request($req),
        $response-status := $resp[1],
        $response-content := $resp[2] return
        
            $response-content
};